=== modified file 'NEWS'
--- NEWS	2008-12-02 05:12:51 +0000
+++ NEWS	2008-12-03 22:29:40 +0000
@@ -15,6 +15,12 @@
       This reverses the change that fixed #294479.
       (Martin Pool, #303856)
 
+    * Properly handle when a revision can be inserted as a delta versus
+      when it needs to be expanded to a fulltext for stacked branches.
+      There was a bug involving merge revisions. As a method to help
+      prevent future difficulties, also make stacked fetches sort
+      topologically. (John Arbash Meinel, #304841)
+
 
 bzr 1.10rc1 2008-11-28
 ----------------------

=== modified file 'bzrlib/knit.py'
--- bzrlib/knit.py	2008-11-27 08:26:50 +0000
+++ bzrlib/knit.py	2008-12-05 14:12:46 +0000
@@ -1333,6 +1333,7 @@
                 adapter = adapter_factory(self)
                 adapters[adapter_key] = adapter
                 return adapter
+        delta_types = set()
         if self._factory.annotated:
             # self is annotated, we need annotated knits to use directly.
             annotated = "annotated-"
@@ -1342,11 +1343,13 @@
             annotated = ""
             convertibles = set(["knit-annotated-ft-gz"])
             if self._max_delta_chain:
+                delta_types.add("knit-annotated-delta-gz")
                 convertibles.add("knit-annotated-delta-gz")
         # The set of types we can cheaply adapt without needing basis texts.
         native_types = set()
         if self._max_delta_chain:
             native_types.add("knit-%sdelta-gz" % annotated)
+            delta_types.add("knit-%sdelta-gz" % annotated)
         native_types.add("knit-%sft-gz" % annotated)
         knit_types = native_types.union(convertibles)
         adapters = {}
@@ -1365,14 +1368,20 @@
         buffered_index_entries = {}
         for record in stream:
             parents = record.parents
+            if record.storage_kind in delta_types:
+                # TODO: eventually the record itself should track
+                #       compression_parent
+                compression_parent = parents[0]
+            else:
+                compression_parent = None
             # Raise an error when a record is missing.
             if record.storage_kind == 'absent':
                 raise RevisionNotPresent([record.key], self)
             elif ((record.storage_kind in knit_types)
-                  and (not parents
+                  and (compression_parent is None
                        or not self._fallback_vfs
-                       or not self._index.missing_keys(parents)
-                       or self.missing_keys(parents))):
+                       or self._index.has_key(compression_parent)
+                       or not self.has_key(compression_parent))):
                 # we can insert the knit record literally if either it has no
                 # compression parent OR we already have its basis in this kvf
                 # OR the basis is not present even in the fallbacks.  In the
@@ -1420,8 +1429,7 @@
                     #
                     # They're required to be physically in this
                     # KnitVersionedFiles, not in a fallback.
-                    compression_parent = parents[0]
-                    if self.missing_keys([compression_parent]):
+                    if not self._index.has_key(compression_parent):
                         pending = buffered_index_entries.setdefault(
                             compression_parent, [])
                         pending.append(index_entry)

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2008-12-02 05:11:07 +0000
+++ bzrlib/repository.py	2008-12-03 22:27:58 +0000
@@ -581,6 +581,7 @@
         self.inventories.add_fallback_versioned_files(repository.inventories)
         self.revisions.add_fallback_versioned_files(repository.revisions)
         self.signatures.add_fallback_versioned_files(repository.signatures)
+        self._fetch_order = 'topological'
 
     def _check_fallback_repository(self, repository):
         """Check that this repository can fallback to repository safely.

=== modified file 'bzrlib/tests/per_repository/test_add_fallback_repository.py'
--- bzrlib/tests/per_repository/test_add_fallback_repository.py	2008-09-04 20:32:04 +0000
+++ bzrlib/tests/per_repository/test_add_fallback_repository.py	2008-12-03 22:27:58 +0000
@@ -45,3 +45,13 @@
         # ... or on the repository directly...
         self.assertEqual({revision_id: (NULL_REVISION,)},
             repo.get_parent_map([revision_id]))
+
+    def test_add_fallback_sets_fetch_order(self):
+        repo = self.make_repository('repo')
+        tree = self.make_branch_and_tree('branch')
+        if not repo._format.supports_external_lookups:
+            self.assertRaises(errors.UnstackableRepositoryFormat,
+                repo.add_fallback_repository, tree.branch.repository)
+            raise TestNotApplicable
+        repo.add_fallback_repository(tree.branch.repository)
+        self.assertEqual('topological', repo._fetch_order)

=== modified file 'bzrlib/tests/test_fetch.py'
--- bzrlib/tests/test_fetch.py	2008-11-25 20:39:57 +0000
+++ bzrlib/tests/test_fetch.py	2008-12-03 22:27:58 +0000
@@ -22,6 +22,7 @@
 from bzrlib import (
     bzrdir,
     errors,
+    osutils,
     merge,
     repository,
     versionedfile,
@@ -458,6 +459,72 @@
             'unordered', False).next()
         self.assertEqual('knit-ft-gz', record.storage_kind)
 
+    def test_fetch_with_fallback_and_merge(self):
+        builder = self.make_branch_builder('source', format='pack-0.92')
+        builder.start_series()
+        # graph
+        #   A
+        #   |\
+        #   B C
+        #   | |
+        #   | D
+        #   | |
+        #   | E
+        #    \|
+        #     F
+        # A & B are present in the base (stacked-on) repository, A-E are
+        # present in the source.
+        # This reproduces bug #304841
+        # We need a large enough inventory that total size of compressed deltas
+        # is shorter than the size of a compressed fulltext. We have to use
+        # random ids because otherwise the inventory fulltext compresses too
+        # well and the deltas get bigger.
+        to_add = [
+            ('add', ('', 'TREE_ROOT', 'directory', None))]
+        for i in xrange(10):
+            fname = 'file%03d' % (i,)
+            fileid = '%s-%s' % (fname, osutils.rand_chars(64))
+            to_add.append(('add', (fname, fileid, 'file', 'content\n')))
+        builder.build_snapshot('A', None, to_add)
+        builder.build_snapshot('B', ['A'], [])
+        builder.build_snapshot('C', ['A'], [])
+        builder.build_snapshot('D', ['C'], [])
+        builder.build_snapshot('E', ['D'], [])
+        builder.build_snapshot('F', ['E', 'B'], [])
+        builder.finish_series()
+        source_branch = builder.get_branch()
+        source_branch.bzrdir.sprout('base', revision_id='B')
+        target_branch = self.make_branch('target', format='1.6')
+        target_branch.set_stacked_on_url('../base')
+        source = source_branch.repository
+        source.lock_read()
+        self.addCleanup(source.unlock)
+        source.inventories = versionedfile.OrderingVersionedFilesDecorator(
+                        source.inventories,
+                        key_priority={('E',): 1, ('D',): 2, ('C',): 4,
+                                      ('F',): 3})
+        # Ensure that the content is yielded in the proper order, and given as
+        # the expected kinds
+        records = [(record.key, record.storage_kind)
+                   for record in source.inventories.get_record_stream(
+                        [('D',), ('C',), ('E',), ('F',)], 'unordered', False)]
+        self.assertEqual([(('E',), 'knit-delta-gz'), (('D',), 'knit-delta-gz'),
+                          (('F',), 'knit-delta-gz'), (('C',), 'knit-delta-gz')],
+                          records)
+
+        target_branch.lock_write()
+        self.addCleanup(target_branch.unlock)
+        target = target_branch.repository
+        target.fetch(source, revision_id='F')
+        # 'C' should be expanded to a fulltext, but D and E should still be
+        # deltas
+        stream = target.inventories.get_record_stream(
+            [('C',), ('D',), ('E',), ('F',)],
+            'unordered', False)
+        kinds = dict((record.key, record.storage_kind) for record in stream)
+        self.assertEqual({('C',): 'knit-ft-gz', ('D',): 'knit-delta-gz',
+                          ('E',): 'knit-delta-gz', ('F',): 'knit-delta-gz'},
+                         kinds)
 
 
 class Test1To2Fetch(TestCaseWithTransport):

=== modified file 'bzrlib/tests/test_versionedfile.py'
--- bzrlib/tests/test_versionedfile.py	2008-07-17 09:13:16 +0000
+++ bzrlib/tests/test_versionedfile.py	2008-12-03 21:05:01 +0000
@@ -2218,3 +2218,58 @@
         record = it.next()
         self.assertEquals("absent", record.storage_kind)
 
+
+class TestOrderingVersionedFilesDecorator(TestCaseWithMemoryTransport):
+
+    def get_ordering_vf(self, key_priority):
+        builder = self.make_branch_builder('test')
+        builder.start_series()
+        builder.build_snapshot('A', None, [
+            ('add', ('', 'TREE_ROOT', 'directory', None))])
+        builder.build_snapshot('B', ['A'], [])
+        builder.build_snapshot('C', ['B'], [])
+        builder.build_snapshot('D', ['C'], [])
+        builder.finish_series()
+        b = builder.get_branch()
+        b.lock_read()
+        self.addCleanup(b.unlock)
+        vf = b.repository.inventories
+        return versionedfile.OrderingVersionedFilesDecorator(vf, key_priority)
+
+    def test_get_empty(self):
+        vf = self.get_ordering_vf({})
+        self.assertEqual([], vf.calls)
+
+    def test_get_record_stream_topological(self):
+        vf = self.get_ordering_vf({('A',): 3, ('B',): 2, ('C',): 4, ('D',): 1})
+        request_keys = [('B',), ('C',), ('D',), ('A',)]
+        keys = [r.key for r in vf.get_record_stream(request_keys,
+                                    'topological', False)]
+        # We should have gotten the keys in topological order
+        self.assertEqual([('A',), ('B',), ('C',), ('D',)], keys)
+        # And recorded that the request was made
+        self.assertEqual([('get_record_stream', request_keys, 'topological',
+                           False)], vf.calls)
+
+    def test_get_record_stream_ordered(self):
+        vf = self.get_ordering_vf({('A',): 3, ('B',): 2, ('C',): 4, ('D',): 1})
+        request_keys = [('B',), ('C',), ('D',), ('A',)]
+        keys = [r.key for r in vf.get_record_stream(request_keys,
+                                   'unordered', False)]
+        # They should be returned based on their priority
+        self.assertEqual([('D',), ('B',), ('A',), ('C',)], keys)
+        # And the request recorded
+        self.assertEqual([('get_record_stream', request_keys, 'unordered',
+                           False)], vf.calls)
+
+    def test_get_record_stream_implicit_order(self):
+        vf = self.get_ordering_vf({('B',): 2, ('D',): 1})
+        request_keys = [('B',), ('C',), ('D',), ('A',)]
+        keys = [r.key for r in vf.get_record_stream(request_keys,
+                                   'unordered', False)]
+        # A and C are not in the map, so they get sorted to the front. A comes
+        # before C alphabetically, so it comes back first
+        self.assertEqual([('A',), ('C',), ('D',), ('B',)], keys)
+        # And the request recorded
+        self.assertEqual([('get_record_stream', request_keys, 'unordered',
+                           False)], vf.calls)

=== modified file 'bzrlib/versionedfile.py'
--- bzrlib/versionedfile.py	2008-11-21 20:20:52 +0000
+++ bzrlib/versionedfile.py	2008-12-03 21:05:01 +0000
@@ -521,7 +521,7 @@
     """
 
     def __init__(self, backing_vf):
-        """Create a RecordingVersionedFileDsecorator decorating backing_vf.
+        """Create a RecordingVersionedFilesDecorator decorating backing_vf.
         
         :param backing_vf: The versioned file to answer all methods.
         """
@@ -562,6 +562,44 @@
         return self._backing_vf.keys()
 
 
+class OrderingVersionedFilesDecorator(RecordingVersionedFilesDecorator):
+    """A VF that records calls, and returns keys in specific order.
+
+    :ivar calls: A list of the calls made; can be reset at any time by
+        assigning [] to it.
+    """
+
+    def __init__(self, backing_vf, key_priority):
+        """Create a RecordingVersionedFilesDecorator decorating backing_vf.
+
+        :param backing_vf: The versioned file to answer all methods.
+        :param key_priority: A dictionary defining what order keys should be
+            returned from an 'unordered' get_record_stream request.
+            Keys with lower priority are returned first, keys not present in
+            the map get an implicit priority of 0, and are returned in
+            lexicographical order.
+        """
+        RecordingVersionedFilesDecorator.__init__(self, backing_vf)
+        self._key_priority = key_priority
+
+    def get_record_stream(self, keys, sort_order, include_delta_closure):
+        self.calls.append(("get_record_stream", list(keys), sort_order,
+            include_delta_closure))
+        if sort_order == 'unordered':
+            def sort_key(key):
+                return (self._key_priority.get(key, 0), key)
+            # Use a defined order by asking for the keys one-by-one from the
+            # backing_vf
+            for key in sorted(keys, key=sort_key):
+                for record in self._backing_vf.get_record_stream([key],
+                                'unordered', include_delta_closure):
+                    yield record
+        else:
+            for record in self._backing_vf.get_record_stream(keys, sort_order,
+                            include_delta_closure):
+                yield record
+
+
 class KeyMapper(object):
     """KeyMappers map between keys and underlying partitioned storage."""
 

